/*  Copyright (c) MIPoPS. All Rights Reserved.
 *
 *  Use of this source code is governed by a BSD-3-Clause license that can
 *  be found in the LICENSE file in the root of the source tree.
 */

//---------------------------------------------------------------------------
#include "Common/ProcessFile.h"
#include "ZenLib/Ztring.h"
#include "Output.h"
#include <iostream>
#include <mutex>
#include <iomanip>
using namespace ZenLib;
//---------------------------------------------------------------------------

//***************************************************************************
// Callback
//***************************************************************************

void __stdcall Event_CallBackFunction(unsigned char* Data_Content, size_t Data_Size, void* UserHandler_Void)
{
    //Retrieving UserHandler
    file*                               UserHandler = (file*)UserHandler_Void;
    struct MediaInfo_Event_Generic*     Event_Generic = (struct MediaInfo_Event_Generic*) Data_Content;
    unsigned char                       ParserID;
    unsigned short                      EventID;
    unsigned char                       EventVersion;

    //Integrity test
    if (Data_Size < 4)
        return; //There is a problem

    //Retrieving EventID
    ParserID = (unsigned char)((Event_Generic->EventCode & 0xFF000000) >> 24);
    EventID = (unsigned short)((Event_Generic->EventCode & 0x00FFFF00) >> 8);
    EventVersion = (unsigned char)(Event_Generic->EventCode & 0x000000FF);
    switch (ParserID)
    {
    case MediaInfo_Parser_DvDif:
        switch (EventID)
        {
        case MediaInfo_Event_DvDif_Analysis_Frame: if (EventVersion == 1 && Data_Size >= sizeof(struct MediaInfo_Event_DvDif_Analysis_Frame_1)) UserHandler->AddFrame((MediaInfo_Event_DvDif_Analysis_Frame_1*)Event_Generic); break;
        case MediaInfo_Event_DvDif_Change: if (EventVersion == 0 && Data_Size >= sizeof(struct MediaInfo_Event_DvDif_Change_0)) UserHandler->AddChange((MediaInfo_Event_DvDif_Change_0*)Event_Generic); break;
        case MediaInfo_Event_Global_Demux: if (EventVersion == 4 && Data_Size >= sizeof(struct MediaInfo_Event_Global_Demux_4)) UserHandler->AddFrame((MediaInfo_Event_Global_Demux_4*)Event_Generic); break;
        }
        break;
    }
}

//***************************************************************************
// Constructor/Destructor
//***************************************************************************

static int Pos = 0;
//---------------------------------------------------------------------------
file::file(const String& FileName)
{
    FilePos = Pos++;

    FrameNumber = 0;

    MI.Option(__T("File_Event_CallBackFunction"), __T("CallBack=memory://") + Ztring::ToZtring((size_t)&Event_CallBackFunction) + __T(";UserHandler=memory://") + Ztring::ToZtring((size_t)this));
    MI.Option(__T("File_DvDif_Analysis"), __T("1"));
    MI.Open(FileName);

    // Filing some info
    FrameRate = Ztring(MI.Get(Stream_Video, 0, __T("FrameRate_Original"))).To_float64();
    if (!FrameRate)
        FrameRate = Ztring(MI.Get(Stream_Video, 0, __T("FrameRate"))).To_float64();
    if (!FrameRate || (FrameRate >= 29.97 && FrameRate <= 29.98))
        FrameRate = double(30 / 1.001); // Default if no frame rate available, or better rounding
}

//---------------------------------------------------------------------------
file::~file()
{
    for (auto& Frame : PerFrame)
    {
        delete[] Frame->Errors;
        delete[] Frame->Video_STA_Errors;
        delete[] Frame->Audio_Data_Errors;
        delete Frame;
    }
    for (auto& Change : PerChange)
    {
        delete Change;
    }
}

//***************************************************************************
// Helpers
//***************************************************************************

//---------------------------------------------------------------------------
void file::AddChange(const MediaInfo_Event_DvDif_Change_0* FrameData)
{
    FrameNumber++; // Event FrameCount is currently wrong

    // Check if there is a change we support
    if (!PerChange.empty())
    {
        const auto Current = PerChange.back();
        if (Current->Width == FrameData->Width
            && Current->Height == FrameData->Height
            && Current->VideoChromaSubsampling == FrameData->VideoChromaSubsampling
            && Current->VideoScanType == FrameData->VideoScanType
            && Current->VideoRatio_N == FrameData->VideoRatio_N
            && Current->VideoRatio_D == FrameData->VideoRatio_D
            && Current->VideoRate_N == FrameData->VideoRate_N
            && Current->VideoRate_D == FrameData->VideoRate_D
            && Current->AudioRate_N == FrameData->AudioRate_N
            && Current->AudioRate_D == FrameData->AudioRate_D
            && Current->AudioChannels == FrameData->AudioChannels
            && Current->AudioBitDepth == FrameData->AudioBitDepth
            && Current->Captions_Flags == FrameData->Captions_Flags)
        {
            return;
        }
    }

    MediaInfo_Event_DvDif_Change_0* ToPush = new MediaInfo_Event_DvDif_Change_0();
    std::memcpy(ToPush, FrameData, sizeof(MediaInfo_Event_DvDif_Change_0));
    ToPush->FrameNumber = FrameNumber - 1; // Event FrameCount is currently wrong
    PerChange.push_back(ToPush);
}

vector<string> FileNames;
string Target;
/*
static const size_t FileCount = 6;
const char* FileNames[] =
{
"C:\\Users\\jerom\\Downloads\\DV\\Compare\\dvgrab_dsr11_pass1_part02.dv",
"C:\\Users\\jerom\\Downloads\\DV\\Compare\\dvgrab_dsr11_pass2_part02.dv",
"C:\\Users\\jerom\\Downloads\\DV\\Compare\\dvgrab_dsr11_pass3_part02.dv",
"C:\\Users\\jerom\\Downloads\\DV\\Compare\\dvgrab_dsr11_pass4_part02.dv",
"C:\\Users\\jerom\\Downloads\\DV\\Compare\\dvgrab_hvr15e_pass1_part2.dv",
"C:\\Users\\jerom\\Downloads\\DV\\Compare\\dvgrab_hvr15e_pass2_part2.dv",
};
const char* Target = "C:\\Users\\jerom\\Downloads\\DV\\Compare\\Target.dv";
//*/
/*
static const size_t FileCount = 6;
const char* FileNames[] =
{
"C:\\Users\\jerom\\Downloads\\DV\\sno_0_162000000\\sno_0_162000000.dv",
"C:\\Users\\jerom\\Downloads\\DV\\sno_0_162000000\\sno_216120000_162000000.dv",
"C:\\Users\\jerom\\Downloads\\DV\\sno_0_162000000\\sno_432600000_162000000.dv",
"C:\\Users\\jerom\\Downloads\\DV\\sno_0_162000000\\sno_636120000_162000000.dv",
"C:\\Users\\jerom\\Downloads\\DV\\sno_0_162000000\\sno_835560000_162000000.dv",
"C:\\Users\\jerom\\Downloads\\DV\\sno_0_162000000\\sno_1046040000_162000000.dv",
};
const char* Target = "C:\\Users\\jerom\\Downloads\\DV\\sno_0_162000000\\Target.dv";
//*/
mutex Mutex;
vector<bool> NOK[100];
static size_t NOK_Pos;
vector<size_t> Counts;
vector<size_t> Counts_Healthy;
vector<size_t> Counts_Healthy_STA;
vector<size_t> Counts_NotHealthy_STA;
vector<size_t> DoNotUseFile;
size_t DoNotUseFiles = 0;
vector<MediaInfo_int8u*> Mergeable[100];
vector<size_t> Mergeable_Counts;
size_t Frames_Good = 0;
size_t Frames_Recovered = 0;
size_t STA_Bad_0 = 0;
size_t STA_Bad_All = 0;
size_t STA_Total = 0;
size_t STA_Fixed = 0;
size_t STA_NotFixed = 0;
FILE* F[100+1];
int FTell[100+1];
uint8_t Buffers[100+1][120000];
size_t PosMin = -1;

//---------------------------------------------------------------------------
void file::AddFrame(const MediaInfo_Event_DvDif_Analysis_Frame_1* FrameData)
{
    MediaInfo_Event_DvDif_Analysis_Frame_1* ToPush = new MediaInfo_Event_DvDif_Analysis_Frame_1();
    std::memcpy(ToPush, FrameData, sizeof(MediaInfo_Event_DvDif_Analysis_Frame_1));
    if (FrameData->Errors)
    {
        size_t SizeToCopy = std::strlen(FrameData->Errors) + 1;
        auto Errors = new char[SizeToCopy];
        std::memcpy(Errors, FrameData->Errors, SizeToCopy);
        ToPush->Errors = Errors;
    }
    if (FrameData->Video_STA_Errors)
    {
        size_t SizeToCopy = FrameData->Video_STA_Errors_Count * sizeof(size_t);
        auto Video_STA_Errors = new size_t[SizeToCopy];
        std::memcpy(Video_STA_Errors, FrameData->Video_STA_Errors, SizeToCopy);
        ToPush->Video_STA_Errors = Video_STA_Errors;
    }
    if (FrameData->Audio_Data_Errors)
    {
        size_t SizeToCopy = FrameData->Audio_Data_Errors_Count * sizeof(size_t);
        auto Audio_Data_Errors = new size_t[SizeToCopy];
        std::memcpy(Audio_Data_Errors, FrameData->Audio_Data_Errors, SizeToCopy);
        ToPush->Audio_Data_Errors = Audio_Data_Errors;
    }
    PerFrame.push_back(ToPush);

    if (Target.empty())
        return;
        
    Mutex.lock();
    auto FileCount = FileNames.size();
    DoNotUseFile.resize(FileCount);
    if (DoNotUseFile[FilePos] || timecode(FrameData->TimeCode).Repeat())
    {
        if (!DoNotUseFile[FilePos])
        {
            DoNotUseFile[FilePos] = true;
            DoNotUseFiles++;
        }
        Mutex.unlock();
        return;
    }
    Counts_Healthy.resize(FileCount);
    Counts_Healthy_STA.resize(FileCount);
    Counts_NotHealthy_STA.resize(FileCount);
    if (!FrameData->Video_STA_Errors)
        Counts_Healthy[FilePos]++;
    for (int i = 0; i < 1500; i++)
    {
        if (FrameData->Mergeable[i] == 1)
            Counts_Healthy_STA[FilePos]++;
        if (FrameData->Mergeable[i] == 2)
            Counts_NotHealthy_STA[FilePos]++;
    }


    NOK[FilePos].push_back(/*FrameData->Errors ||*/ FrameData->Video_STA_Errors || FrameData->Audio_Data_Errors);
    auto Mergeable_Item = new MediaInfo_int8u[1500];
    memcpy(Mergeable_Item, FrameData->Mergeable, 1500);
    Mergeable[FilePos].push_back(Mergeable_Item);
    Counts.resize(FileCount);
    Mergeable_Counts.resize(FileCount);
    PosMin=-1;
    for (size_t i = 0; i < FileCount; i++)
        if (!DoNotUseFile[FilePos] && PosMin > NOK[i].size())
            PosMin = NOK[i].size();

    DisplayLines();
    
    Mutex.unlock();
}

//---------------------------------------------------------------------------
void file::DisplayLines()
{
    auto FileCount = FileNames.size();
    for (; NOK_Pos < PosMin; NOK_Pos++)
    {
        if (!NOK_Pos)
        {
            for (size_t i = 0; i < FileCount; i++)
                F[i] = fopen(FileNames[i].c_str(), "rb");
            F[FileCount] = fopen(Target.c_str(), "wb");
        }
        for (size_t i = 0; i < FileCount; i++)
        {
            if (NOK_Pos < NOK[i].size())
            {
                auto B = fread(Buffers[i], 1, 120000, F[i]);
                if (B != 120000)
                    int A = 0;
            }
        }

        size_t Prefered = -1;
        for (size_t i = 0; i < FileCount; i++)
        {
            if (NOK_Pos < NOK[i].size() && !NOK[i][NOK_Pos])
            {
                Prefered = i;
                break;
            }
        }
        memcpy(Buffers[FileCount], Buffers[0], 120000);
        
        cout << setfill(' ') << setw(6) << NOK_Pos << " ";
        //timecode TC(FrameData->TimeCode);
        //cout << (char)('0' + (TC.TimeInSeconds() % 60) / 10) << (char)('0' + (TC.TimeInSeconds() % 10)) << (TC.DropFrame()? ';' : ':') << (char)('0' + (TC.Frames() % 100) / 10) << (char)('0' + (TC.Frames() % 10)) << ' ';



        if (Prefered != -1)
        {
            cout << Prefered;
            Counts[Prefered]++;
            Frames_Good++;
            if (Prefered)
                memcpy(Buffers[FileCount], Buffers[Prefered], 120000);
        }
        else
            cout << "X";

        cout << "   ("; 
        for (size_t i = 0; i < FileCount; i++)
            if (NOK_Pos < NOK[i].size())
                cout << " " << (NOK[i][NOK_Pos] ? 'P' : ' ');
            else
                cout << " N";

        cout << ")";

        if (Prefered == -1)
        {
            cout << " STA ";
            size_t Mergeable_Counts_Frame[100];
            size_t Issue = 0;
            memset(Mergeable_Counts_Frame, 0, FileCount * sizeof(size_t));
            for (int i = 0; i < 1500; i++)
            {
                bool NoIssue = false;
                for (int j = 0; j < FileCount; j++)
                    if (NOK_Pos < NOK[j].size() && Mergeable[j][NOK_Pos][i] == 1)
                    {
                        if (true)
                        {
                            Mergeable_Counts_Frame[j]++;
                            NoIssue = true;

                            if (j)
                            {
                                memcpy(Buffers[FileCount] + i * 80, Buffers[j] + i * 80, 80);
                                STA_Fixed++;
                            }

                            break;
                        }
                    }
                if (!NoIssue)
                    Issue++;
            }
            for (int j = 0; j < FileCount; j++)
            {
                cout << " " << Mergeable_Counts_Frame[j];
            }
            if (Issue > 150) //TEMP STA only
            {
                Issue -= 150;
                cout << ", and " << Issue << " issues";
                for (int i = 0; i < 1500; i++)
                    if (Mergeable[0][NOK_Pos][i] == 2)
                        STA_Bad_0++;
                STA_Bad_All += Issue;
                STA_Total += 1350;
                STA_NotFixed += Issue;
            }
            else
                Frames_Recovered++;
        }

        fwrite(Buffers[FileCount], 120000, 1, F[FileCount]);
        cout << endl;
    }
}

//---------------------------------------------------------------------------
void file::DisplayEnd()
{
    static int Precision = 6;

    if (Target.empty())
        return;

    auto FileCount = FileNames.size();
    for (size_t i = 0; i < FileCount; i++)
        if (PosMin < NOK[i].size())
            PosMin = NOK[i].size();
    DisplayLines();

    cout << endl;
    FrameNumber = PosMin;
    size_t X = FrameNumber;
    float Percent_Frames_Max = 0;
    float Percent_STA_Max = 0;
    for (size_t i = 0; i < FileCount; i++)
    {
        X -= Counts[i];
        if (Counts[i])
            cout << setfill(' ') << fixed << setw(Precision + 2) << setprecision(Precision - 2) << (float)Counts[i] / FrameNumber * 100 << "% of file " << i << "     used";
        else
        {
            for (size_t j = 0; j < Precision + 2; j++)
                cout << ' ';
            cout << "     file " << i << " not used";
        }
        auto Percent_Frames = (float)Counts_Healthy[i] / FrameNumber * 100;
        auto Percent_STA = (float)Counts_Healthy_STA[i] / (Counts_Healthy_STA[i] + Counts_NotHealthy_STA[i]) * 100;
        cout << " (" << fixed << setw(4) << Counts_Healthy[i] << " = " << setw(Precision + 2) << setprecision(Precision - 2) << Percent_Frames << "% of frames are healthy; " << setw(6) << setprecision(Precision - 2) << Percent_STA << "% of STA are healthy)." << endl;

        if (Percent_Frames_Max < Percent_Frames)
            Percent_Frames_Max = Percent_Frames;
        if (Percent_STA_Max < Percent_STA)
            Percent_STA_Max = Percent_STA;
    }

    cout << endl;
    if (Frames_Good)
        cout << setfill(' ') << fixed << setw(Precision + 2) << setprecision(Precision - 2) << (float)Frames_Good / FrameNumber * 100 << "% of frames are recovered from full frames." << endl;
    if (Frames_Recovered)
        cout << setfill(' ') << fixed << setw(Precision + 2) << setprecision(Precision - 2) << (float)Frames_Recovered / FrameNumber * 100 << "% of frames are fully recovered from video blocks." << endl;
    if (FrameNumber - Frames_Good - Frames_Recovered)
        cout << setfill(' ') << fixed << setw(Precision + 2) << setprecision(Precision - 2) << (float)(FrameNumber - Frames_Good - Frames_Recovered) / FrameNumber * 100 << "% of frames still have issues (in these frames, from " << fixed << setw(Precision + 2) << setprecision(Precision - 2) << (float)STA_Bad_0 / STA_Total * 100 << "% bad in file 0 to " << fixed << setw(Precision + 2) << setprecision(Precision - 2) << (float)(STA_Bad_All) / STA_Total * 100 << "% bad)." << endl;
    if (STA_Fixed)
        cout << STA_Fixed << " STA fixed." << endl;
    if (STA_NotFixed)
        cout << STA_NotFixed << " STA not fixed." << endl;

    cout << endl;
    cout << setfill(' ') << setw(4) << FrameNumber << " frames retrieved." << endl;
    auto Percent_Frames_Final = (float)(Frames_Good + Frames_Recovered) / FrameNumber * 100;
    auto Percent_STA_Final = (float)(1350 * FrameNumber - STA_NotFixed) / 1350 / FrameNumber * 100;
    cout << setfill(' ') << fixed << setw(4) << Frames_Good + Frames_Recovered << " = " << fixed << setw(Precision + 2) << setprecision(Precision - 2) << Percent_Frames_Final << "% of frames are healthy, " << fixed << setw(8) << 1350 * FrameNumber - STA_NotFixed << " = " << fixed << setw(Precision + 2) << setprecision(Precision - 2) << Percent_STA_Final << "% STA are healthy." << endl;
    cout << "     (+" << setfill(' ') << fixed << setw(Precision + 2) << setprecision(Precision - 2) << Percent_Frames_Final - Percent_Frames_Max << ")                                 (+" << fixed << setw(Precision + 2) << setprecision(Precision - 2) << Percent_STA_Final - Percent_STA_Max << ")" << endl;

}

//---------------------------------------------------------------------------
void file::AddFrame(const MediaInfo_Event_Global_Demux_4* FrameData)
{
    if (!FrameData->StreamIDs_Size)
        return;
    auto Dseq = FrameData->StreamIDs[FrameData->StreamIDs_Size-1];
    if ((Dseq & 0xFFFF00) != ((0x2 << 16) | (0x65 << 8))) // identifier with SCT = 0x2, PackType = 0x65, and Dseq
        return;
    Dseq &= 0xFF; // Extract Dseq
    if (Dseq >= 16)
        return;
    if (FrameData->Content_Size != 4)
        return;
    
    // Store
    for (int i = 0; i < 2; i++)
    {
        // Filter unneeded content
        if (FrameData->Content[i * 2] == 0x80 && FrameData->Content[i * 2 + 1] == 0x80)
            return; // Don't store empty content
        if (Dseq && PerFrame_Captions_PerSeq_PerField.size() == 1)
        {
            auto& PerSeq0 = PerFrame_Captions_PerSeq_PerField[0];
            auto& FieldData0 = PerSeq0.FieldData[i];
            if (!FieldData0.empty() && FieldData0.back().StartFrameNumber + FieldData0.back().Captions.size() - 1 == FrameNumber && !FieldData0.back().Captions.back().compare(FrameData->Content + i * 2))
                continue; // Content is same, not stored
        }

        // Store
        if (Dseq >= PerFrame_Captions_PerSeq_PerField.size())
            PerFrame_Captions_PerSeq_PerField.resize(Dseq + 1);
        auto& PerSeq = PerFrame_Captions_PerSeq_PerField[Dseq];
        auto& FieldData = PerSeq.FieldData[i];
        if (FieldData.empty() || FieldData.back().StartFrameNumber + FieldData.back().Captions.size() != FrameNumber)
            FieldData.emplace_back(FrameNumber);
        FieldData.back().Captions.emplace_back(FrameData->Content + i * 2);
    }
}
